# Copyright 2013-2020 Ben Lambert

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for edit_distance.
"""
import unittest

from edit_distance import edit_distance_backpointer
from edit_distance import edit_distance
from edit_distance import highest_match_action
from edit_distance import SequenceMatcher


class TestEditDistance(unittest.TestCase):
    """Class to hold all the tests for this package."""

    def test_edit_distance0(self):
        """Test edit distance between 'ab' and 'dab'."""
        a = ["a", "b"]
        b = ["d", "a", "b"]
        self.assertEqual(edit_distance(a, b), (1, 2))
        bp_expected_result = (
            1,
            2,
            [["insert", 0, 0, 0, 1], ["equal", 0, 1, 1, 2], ["equal", 1, 2, 2, 3]],
        )
        self.assertEqual(edit_distance_backpointer(a, b), bp_expected_result)

    def test_edit_distance1(self):
        """Test edit distance between 'ab' and 'acdab'."""
        a = ["a", "b"]
        b = ["a", "c", "d", "a", "b"]
        self.assertEqual(edit_distance(a, b), (3, 2))
        bp_expected_result = (
            3,
            2,
            [
                ["insert", 0, 0, 0, 1],
                ["insert", 0, 0, 1, 2],
                ["insert", 0, 0, 2, 3],
                ["equal", 0, 1, 3, 4],
                ["equal", 1, 2, 4, 5],
            ],
        )
        self.assertEqual(edit_distance_backpointer(a, b), bp_expected_result)

    def test_edit_distance2(self):
        """Test edit distance for 'hi my name is andy'."""
        a = ["hi", "my", "name", "is", "andy"]
        b = ["hi", "i'm", "my", "name's", "sandy"]
        self.assertEqual(edit_distance(a, b), (4, 1))
        bp_expected_result = (
            4,
            1,
            [
                ["equal", 0, 1, 0, 1],
                ["replace", 1, 2, 1, 2],
                ["replace", 2, 3, 2, 3],
                ["replace", 3, 4, 3, 4],
                ["replace", 4, 5, 4, 5],
            ],
        )
        self.assertEqual(edit_distance_backpointer(a, b), bp_expected_result)

    def test_edit_distance_highest_match(self):
        """Test edit distance for 'hi my name is andy', maximizing matches rather than
        minimizing edits."""
        a = ["hi", "my", "name", "is", "andy"]
        b = ["hi", "i'm", "my", "name's", "sandy"]
        self.assertEqual(
            edit_distance(a, b, action_function=highest_match_action), (4, 2)
        )
        bp_expected_result = (
            4,
            2,
            [
                ["equal", 0, 1, 0, 1],
                ["insert", 1, 1, 1, 2],
                ["equal", 1, 2, 2, 3],
                ["delete", 2, 3, 2, 2],
                ["replace", 3, 4, 3, 4],
                ["replace", 4, 5, 4, 5],
            ],
        )
        self.assertEqual(
            edit_distance_backpointer(a, b, action_function=highest_match_action),
            bp_expected_result,
        )

    def test_edit_distance3(self):
        """Test for 'are you at work now'."""
        a = ["are", "you", "at", "work", "now"]
        b = ["i", "feel", "are", "saying"]
        bp_expected_result = (
            5,
            0,
            [
                ["delete", 0, 1, 0, 0],
                ["replace", 1, 2, 0, 1],
                ["replace", 2, 3, 1, 2],
                ["replace", 3, 4, 2, 3],
                ["replace", 4, 5, 3, 4],
            ],
        )
        self.assertEqual(edit_distance_backpointer(a, b), bp_expected_result)

    def test_edit_distance4(self):
        """Test edit distance against an empty list."""
        a = []
        b = ["a", "c"]
        self.assertEqual(edit_distance(a, b), (2, 0))
        self.assertEqual(edit_distance(b, a), (2, 0))
        self.assertEqual(edit_distance(a, a), (0, 0))

    def test_sequence_matcher(self):
        """Test the sequence matcher."""
        a = ["a", "b"]
        b = ["a", "b", "d", "c"]
        sm = SequenceMatcher(a=a, b=b)
        opcodes = [
            ["equal", 0, 1, 0, 1],
            ["equal", 1, 2, 1, 2],
            ["insert", 2, 2, 2, 3],
            ["insert", 2, 2, 3, 4],
        ]
        self.assertEqual(sm.distance(), 2)
        self.assertEqual(sm.ratio(), 2 / 3)
        self.assertEqual(sm.quick_ratio(), 2 / 3)
        self.assertEqual(sm.real_quick_ratio(), 2 / 3)
        self.assertEqual(sm.distance(), 2)
        # This doesn't return anything, saves the value in the sm cache.
        self.assertTrue(not sm._compute_distance_fast())
        self.assertEqual(sm.get_opcodes(), opcodes)
        self.assertEqual(list(sm.get_matching_blocks()), [[0, 0, 1], [1, 1, 1]])

    def test_sequence_matcher2(self):
        """Test the sequence matcher."""
        a = ["a", "b"]
        b = ["a", "b", "d", "c"]
        sm = SequenceMatcher()
        sm.set_seq1(a)
        sm.set_seq2(b)
        self.assertEqual(sm.distance(), 2)
        sm.set_seqs(b, a)
        self.assertEqual(sm.distance(), 2)

    def test_unsupported(self):
        """
        Test if calling unimplemented methods actually generates an error.
        """
        a = ["a", "b"]
        b = ["a", "b", "d", "c"]
        sm = SequenceMatcher(a=a, b=b)
        with self.assertRaises(NotImplementedError):
            sm.find_longest_match(1, 2, 3, 4)
        with self.assertRaises(NotImplementedError):
            sm.get_grouped_opcodes()

    def test_issue4_simpler(self):
        """Test for error reported here:
        https://github.com/belambert/edit-distance/issues/4"""
        a = ["that", "continuous", "sanction", ":=", "("]
        b = ["continuous", ":=", "(", "sanction", "^"]
        sm = SequenceMatcher(a=a, b=b)
        self.assertEqual(sm.distance(), 4)
        target_opcodes = [
            ["delete", 0, 1, 0, 0],
            ["equal", 1, 2, 0, 1],
            ["delete", 2, 3, 0, 0],
            ["equal", 3, 4, 1, 2],
            ["equal", 4, 5, 2, 3],
            ["insert", 5, 5, 3, 4],
            ["insert", 5, 5, 4, 5],
        ]
        self.assertEqual(sm.get_opcodes(), target_opcodes)

    def test_issue4(self):
        """Test for error reported here:
        https://github.com/belambert/edit-distance/issues/4"""
        a = [
            "that",
            "continuous",
            "sanction",
            ":=",
            "(",
            "flee",
            "U",
            "complain",
            ")",
            "E",
            "attendance",
            "eye",
            "^",
            "flowery",
            "revelation",
            "^",
            "ridiculous",
            "destination",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
        ]  # noqa
        b = [
            "continuous",
            ":=",
            "(",
            "sanction",
            "^",
            "flee",
            "^",
            "attendance",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
            "<EOS>",
        ]  # noqa
        target_opcodes = [
            ["delete", 0, 1, 0, 0],
            ["equal", 1, 2, 0, 1],
            ["delete", 2, 3, 0, 0],
            ["equal", 3, 4, 1, 2],
            ["equal", 4, 5, 2, 3],
            ["insert", 5, 5, 3, 4],
            ["insert", 5, 5, 4, 5],
            ["equal", 5, 6, 5, 6],
            ["replace", 6, 7, 6, 7],
            ["replace", 7, 8, 7, 8],
            ["replace", 8, 9, 8, 9],
            ["replace", 9, 10, 9, 10],
            ["replace", 10, 11, 10, 11],
            ["replace", 11, 12, 11, 12],
            ["replace", 12, 13, 12, 13],
            ["replace", 13, 14, 13, 14],
            ["replace", 14, 15, 14, 15],
            ["replace", 15, 16, 15, 16],
            ["replace", 16, 17, 16, 17],
            ["replace", 17, 18, 17, 18],
            ["equal", 18, 19, 18, 19],
            ["equal", 19, 20, 19, 20],
            ["equal", 20, 21, 20, 21],
            ["equal", 21, 22, 21, 22],
            ["equal", 22, 23, 22, 23],
            ["equal", 23, 24, 23, 24],
            ["equal", 24, 25, 24, 25],
            ["equal", 25, 26, 25, 26],
            ["equal", 26, 27, 26, 27],
            ["equal", 27, 28, 27, 28],
            ["equal", 28, 29, 28, 29],
        ]  # noqa
        sm = SequenceMatcher(a=a, b=b)
        self.assertEqual(sm.distance(), 16)
        self.assertEqual(sm.get_opcodes(), target_opcodes)

    def test_issue13(self):
        sm = SequenceMatcher(a="abc", b="abdc")
        self.assertEqual(
            [
                ["equal", 0, 1, 0, 1],
                ["equal", 1, 2, 1, 2],
                ["insert", 2, 2, 2, 3],
                ["equal", 2, 3, 3, 4],
            ],
            sm.get_opcodes(),
        )



# -*- mode: Python;-*-

# Copyright 2013-2020 Ben Lambert

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Code for computing edit distances.
"""

import sys
import operator
from typing import Optional, Sequence

INSERT: str = "insert"
DELETE: str = "delete"
EQUAL: str = "equal"
REPLACE: str = "replace"


# Cost is basically: was there a match or not.
# The other numbers are cumulative costs and matches.


def lowest_cost_action(ic, dc, sc, im, dm, sm, cost) -> str:
    """Given the following values, choose the action (insertion, deletion,
    or substitution), that results in the lowest cost (ties are broken using
    the 'match' score).  This is used within the dynamic programming algorithm.

    * ic - insertion cost

    * dc - deletion cost

    * sc - substitution cost

    * im - insertion match (score)

    * dm - deletion match (score)

    * sm - substitution match (score)
    """
    best_action = None
    best_match_count = -1
    min_cost = min(ic, dc, sc)
    if min_cost == sc and cost == 0:
        best_action = EQUAL
        best_match_count = sm
    elif min_cost == sc and cost == 1:
        best_action = REPLACE
        best_match_count = sm
    elif min_cost == ic and im > best_match_count:
        best_action = INSERT
        best_match_count = im
    elif min_cost == dc and dm > best_match_count:
        best_action = DELETE
        best_match_count = dm
    else:
        raise Exception("internal error: invalid lowest cost action")
    return best_action


def highest_match_action(ic, dc, sc, im, dm, sm, cost) -> str:
    """Given the following values, choose the action (insertion, deletion, or
    substitution), that results in the highest match score (ties are broken
    using the distance values).  This is used within the dynamic programming
    algorithm.

    * ic - insertion cost

    * dc - deletion cost

    * sc - substitution cost

    * im - insertion match (score)

    * dm - deletion match (score)

    * sm - substitution match (score)
    """
    best_action = None
    lowest_cost = float("inf")
    max_match = max(im, dm, sm)
    if max_match == sm and cost == 0:
        best_action = EQUAL
        lowest_cost = sm
    elif max_match == sm and cost == 1:
        best_action = REPLACE
        lowest_cost = sm
    elif max_match == im and ic < lowest_cost:
        best_action = INSERT
        lowest_cost = ic
    elif max_match == dm and dc < lowest_cost:
        best_action = DELETE
        lowest_cost = dc
    else:
        raise Exception("internal error: invalid highest match action")
    return best_action


class SequenceMatcher(object):
    """
    Similar to the :py:mod:`difflib` :py:class:`~difflib.SequenceMatcher`, but
    uses Levenshtein/edit distance.
    """

    def __init__(
        self,
        a: Optional[Sequence] = None,
        b: Optional[Sequence] = None,
        test=operator.eq,
        action_function=lowest_cost_action,
    ):
        """
        Initialize the object with sequences a and b.  Optionally, one can
        specify a test function that is used to compare sequence elements. This
        defaults to the built in ``eq`` operator (i.e. :py:func:`operator.eq`).
        """
        if a is None:
            a = []
        if b is None:
            b = []
        self.seq1 = a
        self.seq2 = b
        self._reset_object()
        self.action_function = action_function
        self.test = test
        self.dist = None
        self._matches = None
        self.opcodes = None

    def set_seqs(self, a: Sequence, b: Sequence) -> None:
        """Specify two alternative sequences -- reset any cached values."""
        self.set_seq1(a)
        self.set_seq2(b)
        self._reset_object()

    def _reset_object(self) -> None:
        """Clear out the cached values for distance, matches, and opcodes."""
        self.opcodes = None
        self.dist = None
        self._matches = None

    def set_seq1(self, a: Sequence) -> None:
        """Specify a new sequence for sequence 1, resetting cached values."""
        self._reset_object()
        self.seq1 = a

    def set_seq2(self, b: Sequence) -> None:
        """Specify a new sequence for sequence 2, resetting cached values."""
        self._reset_object()
        self.seq2 = b

    def find_longest_match(self, alo, ahi, blo, bhi) -> None:
        """Not implemented!"""
        raise NotImplementedError()

    def get_matching_blocks(self):
        """Similar to :py:meth:`get_opcodes`, but returns only the opcodes that are
        equal and returns them in a somewhat different format
        (i.e. ``(i, j, n)`` )."""
        opcodes = self.get_opcodes()
        match_opcodes = filter(lambda x: x[0] == EQUAL, opcodes)
        return map(
            lambda opcode: [opcode[1], opcode[3], opcode[2] - opcode[1]], match_opcodes
        )

    def get_opcodes(self):
        """Returns a list of opcodes.  Opcodes are the same as defined by
        :py:mod:`difflib`."""
        if not self.opcodes:
            d, m, opcodes = edit_distance_backpointer(
                self.seq1,
                self.seq2,
                action_function=self.action_function,
                test=self.test,
            )
            if self.dist:
                assert d == self.dist
            if self._matches:
                assert m == self._matches
            self.dist = d
            self._matches = m
            self.opcodes = opcodes
        return self.opcodes

    def get_grouped_opcodes(self, n=None):
        """Not implemented!"""
        raise NotImplementedError()

    def ratio(self) -> float:
        """Ratio of matches to the average sequence length."""
        return 2.0 * self.matches() / (len(self.seq1) + len(self.seq2))

    def quick_ratio(self) -> float:
        """Same as :py:meth:`ratio`."""
        return self.ratio()

    def real_quick_ratio(self) -> float:
        """Same as :py:meth:`ratio`."""
        return self.ratio()

    def _compute_distance_fast(self) -> None:
        """Calls edit_distance, and asserts that if we already have values for
        matches and distance, that they match."""
        d, m = edit_distance(
            self.seq1, self.seq2, action_function=self.action_function, test=self.test
        )
        if self.dist:
            assert d == self.dist
        if self._matches:
            assert m == self._matches
        self.dist = d
        self._matches = m

    def distance(self):
        """Returns the edit distance of the two loaded sequences.  This should
        be a little faster than getting the same information from
        :py:meth:`get_opcodes`."""
        if not self.dist:
            self._compute_distance_fast()
        return self.dist

    def matches(self):
        """Returns the number of matches in the alignment of the two sequences.
        This should be a little faster than getting the same information from
        :py:meth:`get_opcodes`."""
        if not self._matches:
            self._compute_distance_fast()
        return self._matches


def edit_distance(
    seq1: Sequence, seq2: Sequence, action_function=lowest_cost_action, test=operator.eq
):
    """
    Computes the edit distance between the two given sequences.  This uses the
    relatively fast method that only constructs two columns of the 2d array
    for edits.  This function actually uses four columns because we track the
    number of matches too.
    """
    m = len(seq1)
    n = len(seq2)
    # Special, easy cases:
    if seq1 == seq2:
        return 0, n
    if m == 0:
        return n, 0
    if n == 0:
        return m, 0
    v0 = [0] * (n + 1)  # The two 'error' columns
    v1 = [0] * (n + 1)
    m0 = [0] * (n + 1)  # The two 'match' columns
    m1 = [0] * (n + 1)
    for i in range(1, n + 1):
        v0[i] = i
    for i in range(1, m + 1):
        v1[0] = i
        for j in range(1, n + 1):
            cost = 0 if test(seq1[i - 1], seq2[j - 1]) else 1
            # The costs
            ins_cost = v1[j - 1] + 1
            del_cost = v0[j] + 1
            sub_cost = v0[j - 1] + cost
            # Match counts
            ins_match = m1[j - 1]
            del_match = m0[j]
            sub_match = m0[j - 1] + int(not cost)

            action = action_function(
                ins_cost, del_cost, sub_cost, ins_match, del_match, sub_match, cost
            )

            if action in [EQUAL, REPLACE]:
                v1[j] = sub_cost
                m1[j] = sub_match
            elif action == INSERT:
                v1[j] = ins_cost
                m1[j] = ins_match
            elif action == DELETE:
                v1[j] = del_cost
                m1[j] = del_match
            else:
                raise Exception("Invalid dynamic programming option returned!")
                # Copy the columns over
        for k in range(n + 1):
            v0[k] = v1[k]
            m0[k] = m1[k]
    return v1[n], m1[n]


def edit_distance_backpointer(
    seq1, seq2, action_function=lowest_cost_action, test=operator.eq
):
    """
    Similar to :py:func:`~edit_distance.edit_distance` except that this
    function keeps backpointers during the search.  This allows us to return
    the opcodes (i.e. the specific edits that were used to change from one
    string to another).  This function contructs the full 2d array for the
    backpointers only.
    """
    m: int = len(seq1)
    n: int = len(seq2)
    # backpointer array:
    bp = [[None for _ in range(n + 1)] for _ in range(m + 1)]

    # Two columns of the distance and match arrays
    d0 = [0] * (n + 1)  # The two 'distance' columns
    d1 = [0] * (n + 1)
    m0 = [0] * (n + 1)  # The two 'match' columns
    m1 = [0] * (n + 1)

    # Fill in the first column
    for i in range(1, n + 1):
        d0[i] = i
        bp[0][i] = INSERT

    for i in range(1, m + 1):
        d1[0] = i
        bp[i][0] = DELETE

        for j in range(1, n + 1):
            cost = 0 if test(seq1[i - 1], seq2[j - 1]) else 1
            # The costs of each action...
            ins_cost = d1[j - 1] + 1  # insertion
            del_cost = d0[j] + 1  # deletion
            sub_cost = d0[j - 1] + cost  # substitution/match

            # The match scores of each action
            ins_match = m1[j - 1]
            del_match = m0[j]
            sub_match = m0[j - 1] + int(not cost)

            action = action_function(
                ins_cost, del_cost, sub_cost, ins_match, del_match, sub_match, cost
            )
            if action == EQUAL:
                d1[j] = sub_cost
                m1[j] = sub_match
                bp[i][j] = EQUAL
            elif action == REPLACE:
                d1[j] = sub_cost
                m1[j] = sub_match
                bp[i][j] = REPLACE
            elif action == INSERT:
                d1[j] = ins_cost
                m1[j] = ins_match
                bp[i][j] = INSERT
            elif action == DELETE:
                d1[j] = del_cost
                m1[j] = del_match
                bp[i][j] = DELETE
            else:
                raise Exception("Invalid dynamic programming action returned!")
        # copy over the columns
        for k in range(n + 1):
            d0[k] = d1[k]
            m0[k] = m1[k]
    opcodes = get_opcodes_from_bp_table(bp)
    return d1[n], m1[n], opcodes


def get_opcodes_from_bp_table(bp):
    """Given a 2d list structure, create opcodes from the best path."""
    x = len(bp) - 1
    y = len(bp[0]) - 1
    opcodes = []
    while x != 0 or y != 0:
        this_bp = bp[x][y]
        if this_bp in [EQUAL, REPLACE]:
            opcodes.append([this_bp, max(x - 1, 0), x, max(y - 1, 0), y])
            x = x - 1
            y = y - 1
        elif this_bp == INSERT:
            opcodes.append([INSERT, x, x, max(y - 1, 0), y])
            y = y - 1
        elif this_bp == DELETE:
            opcodes.append([DELETE, max(x - 1, 0), x, max(y - 1, 0), max(y - 1, 0)])
            x = x - 1
        else:
            raise Exception("Invalid dynamic programming action in BP table!")
    opcodes.reverse()
    return opcodes


def main() -> int:
    """Read two files line-by-line and print edit distances between each pair
    of lines. Will terminate at the end of the shorter of the two files."""

    if len(sys.argv) != 3:
        print(f"Usage: {sys.argv[0]} <file1> <file2>")
        exit(-1)
    file1 = sys.argv[1]
    file2 = sys.argv[2]

    with open(file1) as f1, open(file2) as f2:
        for line1, line2 in zip(f1, f2):
            print(f"Line 1: {line1.strip()}")
            print(f"Line 2: {line2.strip()}")
            dist, _, _ = edit_distance_backpointer(line1.split(), line2.split())
            print(f"Distance: {dist}")
            print("=" * 80)
    return 0


if __name__ == "__main__":
    main()



from setuptools import setup

setup(
    name="edit_distance",
    version="1.0.5",
    author="Ben Lambert",
    author_email="blambert@gmail.com",
    license="LICENSE.txt",
    description="Computing edit distance on arbitrary Python sequences.",
    url="https://github.com/belambert/editdistance",
    keywords=["edit", "distance", "editdistance", "levenshtein"],
    test_suite="test",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
    ],
    entry_points={"console_scripts": ["edit-distance = edit_distance.code:main"]},
)
